@@ -39,7 +39,7 @@
 
 G_BEGIN_DECLS
 
-/* calculate a string size, guarranteed to fit format + args.
+/* calculate a string size, guaranteed to fit format + args.
  */
 gsize	g_printf_string_upper_bound (const gchar* format,
 				     va_list	  args);
@@ -104,7 +104,7 @@ GLogLevelFlags  g_log_set_always_fatal  (GLogLevelFlags  fatal_mask);
 void	_g_log_fallback_handler	(const gchar   *log_domain,
 				 GLogLevelFlags log_level,
 				 const gchar   *message,
-				 gpointer       unused_data);
+				 gpointer       unused_data) G_GNUC_INTERNAL;
 
 
 #ifndef G_LOG_DOMAIN
@@ -123,6 +123,9 @@ void	_g_log_fallback_handler	(const gchar   *log_domain,
 #define g_warning(...)  g_log (G_LOG_DOMAIN,         \
                                G_LOG_LEVEL_WARNING,  \
                                __VA_ARGS__)
+#define g_debug(...)    g_log (G_LOG_DOMAIN,         \
+                               G_LOG_LEVEL_DEBUG,    \
+                               __VA_ARGS__)
 #elif defined(G_HAVE_GNUC_VARARGS)
 #define g_error(format...)      g_log (G_LOG_DOMAIN,         \
                                        G_LOG_LEVEL_ERROR,    \
@@ -136,6 +139,9 @@ void	_g_log_fallback_handler	(const gchar   *log_domain,
 #define g_warning(format...)    g_log (G_LOG_DOMAIN,         \
                                        G_LOG_LEVEL_WARNING,  \
                                        format)
+#define g_debug(format...)      g_log (G_LOG_DOMAIN,         \
+                                       G_LOG_LEVEL_WARNING,  \
+                                       format)
 #else   /* no varargs macros */
 static void
 g_error (const gchar *format,
@@ -173,6 +179,15 @@ g_warning (const gchar *format,
   g_logv (G_LOG_DOMAIN, G_LOG_LEVEL_WARNING, format, args);
   va_end (args);
 }
+static void
+g_debug (const gchar *format,
+         ...)
+{
+  va_list args;
+  va_start (args, format);
+  g_logv (G_LOG_DOMAIN, G_LOG_LEVEL_DEBUG, format, args);
+  va_end (args);
+}
 #endif  /* !__GNUC__ */
 
 typedef void    (*GPrintFunc)           (const gchar    *string);
