@@ -30,6 +30,8 @@
 
 #include "config.h"
 
+#include <ctype.h>
+
 #ifdef HAVE_UNISTD_H
 #include <unistd.h>
 #endif
@@ -73,6 +75,7 @@
 
 #ifdef G_OS_WIN32
 #  include <direct.h>
+#  include <shlobj.h>
 #endif
 
 #ifdef HAVE_CODESET
@@ -96,6 +99,53 @@ const guint glib_micro_version = GLIB_MICRO_VERSION;
 const guint glib_interface_age = GLIB_INTERFACE_AGE;
 const guint glib_binary_age = GLIB_BINARY_AGE;
 
+/**
+ * glib_check_version:
+ * @required_major: the required major version.
+ * @required_minor: the required major version.
+ * @required_micro: the required major version.
+ *
+ * Checks that the GLib library in use is compatible with the
+ * given version. Generally you would pass in the constants
+ * #GLIB_MAJOR_VERSION, #GLIB_MINOR_VERSION, #GLIB_MICRO_VERSION
+ * as the three arguments to this function; that produces
+ * a check that the library in use is compatible with
+ * the version of GLib the application or module was compiled
+ * against.
+ *
+ * Compatibility is defined by two things: first the version
+ * of the running library is newer than the version
+ * @required_major.required_minor.@required_micro. Second
+ * the running library must be binary compatible with the
+ * version @required_major.required_minor.@required_micro
+ * (same major version.)
+ *
+ * Return value: %NULL if the GLib library is compatible with the
+ *   given version, or a string describing the version mismatch.
+ *   The returned string is owned by GLib and must not be modified
+ *   or freed.
+ *
+ * Since: 2.6
+ **/
+const gchar *
+glib_check_version (guint required_major,
+                    guint required_minor,
+                    guint required_micro)
+{
+  gint glib_effective_micro = 100 * GLIB_MINOR_VERSION + GLIB_MICRO_VERSION;
+  gint required_effective_micro = 100 * required_minor + required_micro;
+
+  if (required_major > GLIB_MAJOR_VERSION)
+    return "GLib version too old (major mismatch)";
+  if (required_major < GLIB_MAJOR_VERSION)
+    return "GLib version too new (major mismatch)";
+  if (required_effective_micro < glib_effective_micro - GLIB_BINARY_AGE)
+    return "GLib version too new (micro mismatch)";
+  if (required_effective_micro > glib_effective_micro)
+    return "GLib version too old (micro mismatch)";
+  return NULL;
+}
+
 #if !defined (HAVE_MEMMOVE) && !defined (HAVE_WORKING_BCOPY)
 void 
 g_memmove (gpointer dest, gconstpointer src, gulong len)
@@ -774,7 +824,13 @@ g_setenv (const gchar *variable,
 /* According to the Single Unix Specification, environ is not in 
  * any system header, although unistd.h often declares it.
  */
+#  ifndef _MSC_VER
+/*
+ * Win32 - at least msvc headers declare it so let's avoid
+ *   warning C4273: '__p__environ' : inconsistent dll linkage.  dllexport assumed.
+ */
 extern char **environ;
+#  endif
 #endif
            
 /**
@@ -831,6 +887,51 @@ static	gchar	*g_user_name = NULL;
 static	gchar	*g_real_name = NULL;
 static	gchar	*g_home_dir = NULL;
 
+static  gchar   *g_user_data_dir = NULL;
+static  gchar  **g_system_data_dirs = NULL;
+static  gchar   *g_user_cache_dir = NULL;
+static  gchar   *g_user_config_dir = NULL;
+static  gchar  **g_system_config_dirs = NULL;
+
+#ifdef G_OS_WIN32
+
+static gchar *
+get_special_folder (int csidl)
+{
+  union {
+    char c[MAX_PATH+1];
+    wchar_t wc[MAX_PATH+1];
+  } path;
+  HRESULT hr;
+  LPITEMIDLIST pidl = NULL;
+  BOOL b;
+  gchar *retval = NULL;
+
+  hr = SHGetSpecialFolderLocation (NULL, csidl, &pidl);
+  if (hr == S_OK)
+    {
+      if (G_WIN32_HAVE_WIDECHAR_API ())
+	{
+	  b = SHGetPathFromIDListW (pidl, path.wc);
+	  if (b)
+	    retval = g_utf16_to_utf8 (path.wc, -1, NULL, NULL, NULL);
+	}
+      else
+	{
+	  b = SHGetPathFromIDListA (pidl, path.c);
+	  if (b)
+	    retval = g_locale_to_utf8 (path.c, -1, NULL, NULL, NULL);
+	}
+      CoTaskMemFree (pidl);
+    }
+  if (retval == NULL)
+    return "C:\\";
+  else
+    return retval;
+}
+
+#endif
+
 /* HOLDS: g_utils_global_lock */
 static void
 g_get_any_init (void)
@@ -867,7 +968,13 @@ g_get_any_init (void)
       /* We check $HOME first for Win32, though it is a last resort for Unix
        * where we prefer the results of getpwuid().
        */
-      g_home_dir = g_strdup (g_getenv ("HOME"));
+      {
+	gchar *home = g_getenv ("HOME");
+      
+	/* Only believe HOME if it is an absolute path and exists */
+	if (home && g_path_is_absolute (home) && g_file_test (home, G_FILE_TEST_IS_DIR))
+	  g_home_dir = g_strdup (home);
+      }
       
       /* In case HOME is Unix-style (it happens), convert it to
        * Windows style.
@@ -886,6 +993,9 @@ g_get_any_init (void)
 	    g_home_dir = g_strdup (g_getenv ("USERPROFILE"));
 	}
 
+      if (!g_home_dir)
+	g_home_dir = get_special_folder (CSIDL_PROFILE);
+
       if (!g_home_dir)
 	{
 	  /* At least at some time, HOMEDRIVE and HOMEPATH were used
@@ -1179,6 +1289,237 @@ g_set_application_name (const gchar *application_name)
     g_warning ("g_set_application() name called multiple times");
 }
 
+/**
+ * g_get_user_data_dir:
+ * 
+ * Returns a base directory in which to access application data such
+ * as icons that is customized for a particular user.  
+ *
+ * On Unix platforms this is determined using the mechanisms described in
+ * the <ulink url="http://www.freedesktop.org/Standards/basedir-spec">
+ * XDG Base Directory Specification</ulink>
+ * 
+ * Return value: a string owned by GLib that must not be modified 
+ *               or freed.
+ * Since: 2.6
+ **/
+G_CONST_RETURN gchar*
+g_get_user_data_dir (void)
+{
+  gchar *data_dir;  
+
+  G_LOCK (g_utils_global);
+
+  if (!g_user_data_dir)
+    {
+#ifdef G_OS_WIN32
+      data_dir = get_special_folder (CSIDL_PERSONAL);
+#else
+      data_dir = (gchar *) g_getenv ("XDG_DATA_HOME");
+
+      if (data_dir && data_dir[0])
+        data_dir = g_strdup (data_dir);
+      else
+	{
+	  if (!g_tmp_dir)
+	    g_get_any_init ();
+
+	  data_dir = g_build_filename (g_home_dir, ".local", 
+				       "share", NULL);
+	}
+#endif
+      g_user_data_dir = data_dir;
+    }
+  else
+    data_dir = g_user_data_dir;
+
+  G_UNLOCK (g_utils_global);
+
+  return data_dir;
+}
+
+/**
+ * g_get_user_config_dir:
+ * 
+ * Returns a base directory in which to store user-specific application 
+ * configuration information such as user preferences and settings. 
+ *
+ * On Unix platforms this is determined using the mechanisms described in
+ * the <ulink url="http://www.freedesktop.org/Standards/basedir-spec">
+ * XDG Base Directory Specification</ulink>
+ * 
+ * Return value: a string owned by GLib that must not be modified 
+ *               or freed.
+ * Since: 2.6
+ **/
+G_CONST_RETURN gchar*
+g_get_user_config_dir (void)
+{
+  gchar *config_dir;  
+
+  G_LOCK (g_utils_global);
+
+  if (!g_user_config_dir)
+    {
+#ifdef G_OS_WIN32
+      config_dir = get_special_folder (CSIDL_APPDATA);
+#else
+      config_dir = (gchar *) g_getenv ("XDG_CONFIG_HOME");
+      
+      if (config_dir && config_dir[0])
+	config_dir = g_strdup (config_dir);
+      else
+	{
+	  if (!g_tmp_dir)
+	    g_get_any_init ();
+	  
+	  config_dir = g_build_filename (g_home_dir, ".config", NULL);
+	}
+#endif      
+      g_user_config_dir = config_dir;
+    }
+  else
+    config_dir = g_user_config_dir;
+
+  G_UNLOCK (g_utils_global);
+
+  return config_dir;
+}
+
+/**
+ * g_get_user_cache_dir:
+ * 
+ * Returns a base directory in which to store non-essential, cached
+ * data specific to particular user.
+ *
+ * On Unix platforms this is determined using the mechanisms described in
+ * the <ulink url="http://www.freedesktop.org/Standards/basedir-spec">
+ * XDG Base Directory Specification</ulink>
+ * 
+ * Return value: a string owned by GLib that must not be modified 
+ *               or freed.
+ * Since: 2.6
+ **/
+G_CONST_RETURN gchar*
+g_get_user_cache_dir (void)
+{
+  gchar *cache_dir;  
+
+  G_LOCK (g_utils_global);
+
+  if (!g_user_cache_dir)
+    {
+#ifdef G_OS_WIN32
+      cache_dir = get_special_folder (CSIDL_INTERNET_CACHE); /* XXX correct? */
+#else
+      cache_dir = (gchar *) g_getenv ("XDG_CACHE_HOME");
+
+      if (cache_dir && cache_dir[0])
+          cache_dir = g_strdup (cache_dir);
+      else
+	{
+	  if (!g_tmp_dir)
+	    g_get_any_init ();
+
+          cache_dir = g_build_filename (g_home_dir, ".cache", NULL);
+	}
+#endif
+      g_user_cache_dir = cache_dir;
+    }
+  else
+    cache_dir = g_user_cache_dir;
+
+  G_UNLOCK (g_utils_global);
+
+  return cache_dir;
+}
+
+/**
+ * g_get_system_data_dirs:
+ * 
+ * Returns an ordered list of base directories in which to access 
+ * system-wide application data.
+ *
+ * On Unix platforms this is determined using the mechanisms described in
+ * the <ulink url="http://www.freedesktop.org/Standards/basedir-spec">
+ * XDG Base Directory Specification</ulink>
+ * 
+ * Return value: a %NULL-terminated array of strings owned by GLib that must 
+ *               not be modified or freed.
+ * Since: 2.6
+ **/
+G_CONST_RETURN gchar * G_CONST_RETURN * 
+g_get_system_data_dirs (void)
+{
+  gchar *data_dirs, **data_dir_vector;
+
+  G_LOCK (g_utils_global);
+
+  if (!g_system_data_dirs)
+    {
+#ifdef G_OS_WIN32
+      data_dirs = g_strconcat (get_special_folder (CSIDL_COMMON_APPDATA),
+			       G_SEARCHPATH_SEPARATOR_S,
+			       get_special_folder (CSIDL_COMMON_DOCUMENTS),
+			       NULL);
+#else
+      data_dirs = (gchar *) g_getenv ("XDG_DATA_DIRS");
+
+      if (!data_dirs || !data_dirs[0])
+          data_dirs = "/usr/local/share/:/usr/share/";
+#endif
+      data_dir_vector = g_strsplit (data_dirs, G_SEARCHPATH_SEPARATOR_S, 0);
+
+      g_system_data_dirs = data_dir_vector;
+    }
+  else
+    data_dir_vector = g_system_data_dirs;
+
+  G_UNLOCK (g_utils_global);
+
+  return (G_CONST_RETURN gchar * G_CONST_RETURN *) data_dir_vector;
+}
+
+/**
+ * g_get_system_config_dirs:
+ * 
+ * Returns an ordered list of base directories in which to access 
+ * system-wide configuration information.
+ *
+ * On Unix platforms this is determined using the mechanisms described in
+ * the <ulink url="http://www.freedesktop.org/Standards/basedir-spec">
+ * XDG Base Directory Specification</ulink>
+ * 
+ * Return value: a %NULL-terminated array of strings owned by GLib that must 
+ *               not be modified or freed.
+ * Since: 2.6
+ **/
+G_CONST_RETURN gchar * G_CONST_RETURN *
+g_get_system_config_dirs (void)
+{
+  gchar *conf_dirs, **conf_dir_vector;
+
+  G_LOCK (g_utils_global);
+
+  if (!g_system_config_dirs)
+    {
+#ifdef G_OS_WIN32
+      conf_dirs = get_special_folder (CSIDL_COMMON_APPDATA);
+#else
+      conf_dirs = (gchar *) g_getenv ("XDG_CONFIG_DIRS");
+
+      if (!conf_dirs || !conf_dirs[0])
+          conf_dirs = "/etc/xdg";
+#endif
+      conf_dir_vector = g_strsplit (conf_dirs, G_SEARCHPATH_SEPARATOR_S, 0);
+    }
+  else
+    conf_dir_vector = g_system_config_dirs;
+  G_UNLOCK (g_utils_global);
+
+  return (G_CONST_RETURN gchar * G_CONST_RETURN *) conf_dir_vector;
+}
+
 guint
 g_direct_hash (gconstpointer v)
 {
@@ -1230,18 +1571,11 @@ g_nullify_pointer (gpointer *nullify_location)
 gchar *
 g_get_codeset (void)
 {
-#ifdef HAVE_CODESET  
-  char *result = nl_langinfo (CODESET);
-  return g_strdup (result);
-#else
-#ifdef G_PLATFORM_WIN32
-  return g_strdup_printf ("CP%d", GetACP ());
-#else
-  /* FIXME: Do something more intelligent based on setlocale (LC_CTYPE, NULL)
-   */
-  return g_strdup ("ISO-8859-1");
-#endif
-#endif
+  const gchar *charset;
+
+  g_get_charset (&charset);
+
+  return g_strdup (charset);
 }
 
 #ifdef ENABLE_NLS
